/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    misc.c - Miscellaneous stuff!
  ***************************************/


#include <errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <netdb.h>

#include "common.h"

/* ------------------------------------------------------------- **
** Listens on socket. If port==0 then picks a port at random (10000-20000)
** There will still be TCP_WAIT problems occasionally.
** ------------------------------------------------------------- */
int listen_on_socket(address_info *listen_address)
{
	int sockfd, len;
	struct sockaddr_in address;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	if(listen_address->port!=0){
		address.sin_family      = AF_INET;
		address.sin_addr.s_addr = listen_address->addr;
		address.sin_port        = listen_address->port;
		len = sizeof(address);
		
		if(bind(sockfd, (struct sockaddr *)&address, len)) {
			perror("bind");
			exit(1);
		}
		if(listen(sockfd, 5)) exit(1);
	}else {
		int i,j;
		srand(time(NULL));

		address.sin_family      = AF_INET;
		address.sin_addr.s_addr = listen_address->addr;
		for(i=0; i<1000; i++){
			j=(rand() % 10000) + 10000;
			address.sin_port = htons(j);
			
			if(!bind(sockfd, (struct sockaddr *)&address, sizeof(address))) break;
			if(errno != EADDRINUSE) i=1002;
		}
		if(i>1001) {
			close(sockfd);
			return(-1);
		}
		if(listen(sockfd, 5)){
                        close(sockfd);
                        return(-1);
                }
		listen_address->port=htons(j);
	}

	return (sockfd);
}

int connect_to_socket(address_info connect_address)
{
	int sockfd, len, tmp;
	struct sockaddr_in address;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	address.sin_family      = AF_INET;
	address.sin_addr.s_addr = connect_address.addr;
	address.sin_port        = connect_address.port;
	len = sizeof(address);

	tmp = connect(sockfd, (struct sockaddr *)&address, len);

	if(tmp==0) return(sockfd);

	close(sockfd);
	return(-1);
}


/* ------------------------------------------------------------- **
** Convert a comma separted values address:port to address_info 
** in network order
** ------------------------------------------------------------- */
address_info com2n(int a1,int a2,int a3,int a4,int p1,int p2)
{
	address_info ret;

        ret.addr = htonl(a4+ (a3<<8) + (a2<<16) + (a1<<24));
        ret.port = htons( (p1<<8) + p2);
	
	return(ret);
}


/* ------------------------------------------------------------- **
** Convert network order address to comma separated values
** ------------------------------------------------------------- */
void n2com(address_info address, int* a1,int* a2,int* a3,int* a4,
	   int* p1,int* p2)
{
	address.addr=ntohl(address.addr);
        address.port=ntohs(address.port);
        *a1=(address.addr & 0xFF000000) >>24;
        *a2=(address.addr & 0x00FF0000) >>16;
        *a3=(address.addr & 0x0000FF00) >>8;
        *a4=(address.addr & 0x000000FF);
        *p1=(address.port & 0xFF00) >>8;
        *p2=(address.port & 0x00FF);
}

/* ------------------------------------------------------------- **
** Extract a comma delimited address/port from buf. buf is NULL
** terminated, but otherwise unchecked user input --- make sure we
** don't buffer overflow here!
** ------------------------------------------------------------- */
address_info extract_address(const char *buf)
{ 
	int i, a[6]; 
	address_info tmp;

	tmp.addr=tmp.port=0;

	for(i=0;i<6;i++){
		if(buf==NULL) return(tmp);
		buf=strpbrk(buf, "0123456789"); 
		if(buf==NULL) return(tmp);
		a[i]=atoi(buf);
		buf=strpbrk(buf, ",");
	}

        tmp=com2n(a[0],a[1],a[2],a[3],a[4],a[5]);
	return(tmp);
}

void write_log(loglevel priority, char * msg, ...)
{
	char buf[MAX_LINE_LEN],buf2[MAX_LINE_LEN];
	va_list argptr;

	va_start(argptr, msg);
	vsnprintf(buf, MAX_LINE_LEN-10, msg, argptr);
	va_end(argptr);
	buf[MAX_LINE_LEN-9]=0;

	switch(priority){
	case ERROR:
	case ATTACK:
		syslog(LOG_ERR|LOG_DAEMON, "%s\n", buf);
		break;
	default:
		syslog(LOG_NOTICE|LOG_DAEMON, "%s\n", buf);
	}

	sprintf(buf2,"frox[%d] %s\n", getpid(), buf);
	write(2, buf2, strlen(buf2));
}

/* ------------------------------------------------------------- **
** Return hostname of address, or failing that the IP as a string
** ------------------------------------------------------------- */
char *addr2name(u_int32_t address)
{
	struct in_addr addr;
	struct hostent *hostinfo;
	static char buf[MAX_LINE_LEN];

	addr.s_addr=address;

	if( (hostinfo=gethostbyaddr((char *)&addr, sizeof(addr), AF_INET))!=NULL)
		strncpy(buf, hostinfo->h_name, MAX_LINE_LEN-1);

	else strncpy(buf, inet_ntoa(addr), MAX_LINE_LEN-1);

	buf[MAX_LINE_LEN-1]=0;
	return(buf);
}

/* ------------------------------------------------------------- **
** Deal with session quit conditions.
** ------------------------------------------------------------- */
void quit(exitreason reason)
{
	switch(reason){
	case TIMEOUT:
		write_log(INFO, "Select timed out.");
		closeandexit(0);
	case SERVER_CLOSE:
		write_log(INFO, "Server closed the control connection.");
		closeandexit(0);
	case CLIENT_CLOSE:
		write_log(INFO, "Client closed the control connection.");
		closeandexit(0);

	case NO_SERVER_DATA:
		write_log(ERROR, "  Unable to connect to server data port");
		closeandexit(-1);
	case NO_CLIENT_DATA:
		write_log(ERROR, "  Unable to connect to client data port");
		closeandexit(-1);
	case MALLOC:
		write_log(ERROR, "  malloc failed");
		closeandexit(-1);
	case CLIENT_RUBBISH:
		write_log(ATTACK, "Client is sending us a badly formed control stream.");
		closeandexit(-1);
	case SERVER_RUBBISH:
		write_log(ATTACK, "Server is sending us a badly formed control stream.");
		closeandexit(-1);
	}
}		


void closeandexit(int status)
{
	close(info->client_control.fd);
	close(info->client_data.fd);
	close(info->server_control.fd);
	close(info->server_data.fd);
	close(info->listen);
	free(info);
	write_log(INFO, "Closing session");
	exit(status);
}
